/*
 * Copyright (C) 2013 - 2015  Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include "colorbutton.h"
#include <QColorDialog>

namespace Fm {

ColorButton::ColorButton(QWidget* parent): QPushButton(parent) {
    connect(this, &QPushButton::clicked, this, &ColorButton::onClicked);
}

ColorButton::~ColorButton() {

}

void ColorButton::onClicked() {
    QColorDialog dlg(color_);
    if(dlg.exec() == QDialog::Accepted) {
        setColor(dlg.selectedColor());
    }
}

void ColorButton::setColor(const QColor& color) {
    if(color != color_) {
        color_ = color;
        // use qss instead of QPalette to set the background color
        // otherwise, this won't work when using the gtk style.
        QString style = QStringLiteral("QPushButton{background-color:%1;}").arg(color.name());
        setStyleSheet(style);
        Q_EMIT changed();
    }
}


} // namespace Fm
